/*-
 * Copyright (C) 2006 Erik Larsson
 * 
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */

package org.catacombae.jlink;

import sf.jlink.*;
import org.catacombae.rarx.*;
import java.io.File;
import java.io.RandomAccessFile;
import java.io.IOException;
import java.util.Vector;
import java.util.Hashtable;
import java.util.regex.Pattern;

public class RARFile {
    /*
     * To speed up things when getting a new stream, I only create one RARFileEntryStream
     * object per RAR entry. Otherwise the entire RAR file would have to be re-parsed every
     * time getFile is called, and that's often.
     * This is a vulnerable method though. Concurrent access is totally out of the question,
     * and the caller must always ensure that it closes the stream after using it.
     * (If it doesn't, the subsequent call to getFile will close it. But it will print an
     *  error message, since it's an indication of possible concurrent access...)
     */
    public static final String continuedRARPattern = ".+\\.[P|p]art(\\d*[^1]|(0*[^0&&\\d]+0*)+1)\\.rar";//.+\\.[P|p]art(\\d*[^1]|[^0]+1)\\.rar";
    public static final String FILE_SEPARATOR = System.getProperty("file.separator");
    public static Hashtable<String, RARFileEntryStream> streams = new Hashtable<String, RARFileEntryStream>();
    private static boolean debug = false;
    
    public static RARFileEntryStream getFile(String rootPath, String requestPath) {
	try {
	    if(debug) System.out.println("getFile(" + rootPath + ", " + requestPath);
	    //String subpath = path.substring(0, path.lastIndexOf(FileSystem.pathSeparatorChar));
	    //String subpath = path.substring(rootPath.length);
	    final String rawArchivePath = pathWithinRARFile(rootPath, requestPath);
	    final String filePath = rootPath + convertToSystemSeparators(requestPath.substring(0, requestPath.length()-rawArchivePath.length()));
	    final String archivePath = normalizePath(convertToRARSeparators(rawArchivePath));
	    final String hashKey = filePath+"\\:/"+archivePath;
	
	    if(debug) System.out.println(" rawArchivePath=" + rawArchivePath);
	    if(debug) System.out.println(" archivePath=" + archivePath);
	    if(debug) System.out.println(" filePath=" + filePath);
	    if(debug) System.out.println(" hashKey=" + hashKey);
// 	path = path.substring();
// 	int firstPathSeparatorIndex = subpath.indexOf(FileSystem.pathSeparatorChar);
// 	String rarFilename;
// 	String pathInFile;
// 	if(firstPathSeparatorIndex == -1) {
// 	    rarFilename = rootPath + "/" + subpath;
// 	    pathInFile = "";
// 	}
// 	else {
// 	    rarFilename = rootPath + "/" + subpath.substring(0, firstPathSeparatorIndex);
// 	    pathInFile = subpath.substring(rarFilename.length+1);
// 	}

	    if(isEntryRARFilename(filePath)) { //filePath.endsWith(".rar") && !Pattern.matches(continuedRARPattern, filePath)) {
		RARFileEntryStream existingStream = streams.get(hashKey);
		if(existingStream != null) {
		    if(!existingStream.isClosed()) {
			System.err.println("RARFile.getFile: WARNING - The stream was not previously closed!");
			existingStream.close();
		    }
		    existingStream.reopen();
		    return existingStream;
		}
		else {
		    File rarFile = new File(filePath);
		    //String entryName = path.substring(path.lastIndexOf("/")+1, path.length());
		    long position = org.catacombae.rarx.RARFile.findEntry(archivePath, new RandomAccessFile(rarFile, "r"));
		    if(position != -1) {
			RARFileEntryStream newStream = new RARFileEntryStream(rarFile, position);
			streams.put(hashKey, newStream);
			return newStream;
		    }
		    else
			return null;
		}
	    }
	    else
		return null;
	} catch(IOException ioe) { return null; }
    }

    public static long getSizeOfEntry(String rootPath, String requestPath) {
	try {
	    if(debug) System.out.println("getSizeOfEntry(" + rootPath + ", " + requestPath);
	    final String rawArchivePath = pathWithinRARFile(rootPath, requestPath);
	    if(rawArchivePath == null)
		return -1;
	    final String filePath = rootPath + convertToSystemSeparators(requestPath.substring(0, requestPath.length()-rawArchivePath.length()));
	    final String archivePath = normalizePath(convertToRARSeparators(rawArchivePath));
	    final String hashKey = filePath+"\\:/"+archivePath;
	
	    if(debug) System.out.println(" rawArchivePath=" + rawArchivePath);
	    if(debug) System.out.println(" archivePath=" + archivePath);
	    if(debug) System.out.println(" filePath=" + filePath);
	    if(debug) System.out.println(" hashKey=" + hashKey);
	    if(isEntryRARFilename(filePath)) { //filePath.endsWith(".rar") && !Pattern.matches(continuedRARPattern, filePath)) {
		RARFileEntryStream existingStream = streams.get(hashKey);
		if(existingStream != null) {
		    return existingStream.length();
		}
		else {
		    File rarFile = new File(filePath);
		    //String entryName = path.substring(path.lastIndexOf("/")+1, path.length());
		    long position = org.catacombae.rarx.RARFile.findEntry(archivePath, new RandomAccessFile(rarFile, "r"));
		    if(position != -1) {
			RARFileEntryStream newStream = new RARFileEntryStream(rarFile, position);
			streams.put(hashKey, newStream);
			newStream.close();
			return newStream.length();
		    }
		    else
			return -1;
		}
	    }
	    else
		return -1;
	} catch(IOException ioe) { return -1; }
	/*
                String name = file[i].getName();
                if(file[i].isDirectory() || (name.endsWith(".pls") && supportPLS) || (name.endsWith(".m3u") && supportM3U) || (name.endsWith(".rar") && supportRAR))
                    name += FileSystem.pathSeparatorChar;
	*/
    }
    
    public static Vector list(String requestPath, String pathInsideArchive) {
	//String 
	String currentRar = convertToSystemSeparators(requestPath.substring(0, requestPath.length()-pathInsideArchive.length()));
	String correctPathInFile = normalizePath(convertToRARSeparators(pathInsideArchive));
	if(correctPathInFile.length() > 0)
	    correctPathInFile += NewFileHeader.PATH_SEPARATOR;
	if(debug) System.out.println("Getting contents of file " + currentRar + "...");
	if(debug) System.out.println(" correctPathInFile=\"" + correctPathInFile + "\"");
	Vector<String> ret = new Vector<String>();
	try {
	    RARFileEntry[] entries = ListFilesInArchive.listFiles(new File(currentRar));
	    for(RARFileEntry entry : entries) {
		NewFileHeader nfh = entry.getHeader(0);
		//System.out.println((nfh.hasIncomingData()?"<--":"   ") + nfh.getFilenameAsString() + (nfh.hasOutgoingData()?"-->":""));
		String filename = nfh.getFilenameAsString();
		if(filename.startsWith(correctPathInFile)) {
		    if(debug) System.out.println("  Found a possible match: \"" + filename + "\"");
		    filename = filename.substring(correctPathInFile.length());
		    if(debug) System.out.println("  Trimmed: \"" + filename + "\"");
		    if(filename.indexOf(NewFileHeader.PATH_SEPARATOR) == -1 &&
		       nfh.getMethod() == RARHeader.METHOD_UNCOMPRESSED &&
		       !(nfh instanceof CommentHeader)) {
			if(nfh.getFileAttributesStructured().isDirectory())
			    filename += FileSystem.pathSeparatorChar;
			ret.add(filename);
		    }
		}
	    }
	} catch(InvalidDataException ide) {
	    if(debug) System.out.println("Invalid RAR-file. Message: " + ide.getMessage());
	    if(debug) System.out.println("File: \"" + currentRar + "\"");
	} catch(Exception e) {
	    e.printStackTrace();
	}
	return ret;
    }

    public static String pathWithinRARFile(String rootPath, String requestPath) {
	String pathWithinFile = "";
	String rarFilePath = requestPath;
	int indexOfSeparator = rarFilePath.length();
	while(indexOfSeparator != -1) {
	    rarFilePath = rarFilePath.substring(0, indexOfSeparator);
	    pathWithinFile = requestPath.substring(rarFilePath.length());
	    if(debug) System.out.println("Testing if \"" + rarFilePath + "\" is valid under \"" + rootPath + "\"");
	    File f = new File(rootPath + convertToSystemSeparators(rarFilePath));
	    if(f.exists()) {
		if(debug) System.out.println("Found file \"" + rootPath + "\" + \"" + convertToSystemSeparators(rarFilePath) + "\"");
		if(isEntryRARFilename(f.getName())) //f.getName().endsWith(".rar") && !Pattern.matches(continuedRARPattern, f.getName()))
		    return pathWithinFile;
		else
		    return null;
	    }
	    indexOfSeparator = rarFilePath.lastIndexOf(FileSystem.pathSeparatorChar);
	}
	return null;
    }

    private static String convertToSystemSeparators(String s) {
	String result = "";
	for(int i = 0; i < s.length(); ++i) {
	    String currentChar = s.substring(i, i+1); //I'm scared of char[]s...
	    if(currentChar.equals("" + FileSystem.pathSeparatorChar))
		result += FILE_SEPARATOR;
	    else
		result += currentChar;
	}
	return result;
    }
    private static String convertToRARSeparators(String s) {
	String result = "";
	for(int i = 0; i < s.length(); ++i) {
	    String currentChar = s.substring(i, i+1); //I'm scared of char[]s...
	    if(currentChar.equals("" + FileSystem.pathSeparatorChar))
		result += NewFileHeader.PATH_SEPARATOR;
	    else
		result += currentChar;
	}
	return result;
    }
    
    private static String normalizePath(String path) {
        //path = path.trim();
        if(path.endsWith(NewFileHeader.PATH_SEPARATOR))
	    //path += NewFileHeader.PATH_SEPARATOR;
	    path = path.substring(0, path.length()-1);
        if(path.startsWith(NewFileHeader.PATH_SEPARATOR))
            path = path.substring(1);
	
        return path;
    }
    
    /**
     * Determines if the name of this file matches the pattern for a RAR-archive
     * which is either composed of one single file, or is the first file in a
     * split archive.
     */
    public static boolean isEntryRARFilename(String filename) {
	return (filename.endsWith(".rar") && !Pattern.matches(continuedRARPattern, filename)) || filename.endsWith(".001");
    }

    /** Test code. */
    public static void main(String[] args) {
	System.out.println("args.length=" + args.length);
	for(String arg : args)
	    System.out.println("  " + arg);
	    
	System.out.println(pathWithinRARFile(args[0], args[1]));
    }
}
